#pragma once

#include <functional>

// BaseClass
#include "MemoryItem.h"

// ChildType
#include "CoreSegmentBase.h"

// ArmarX
#include <ArmarXCore/core/time/TimeUtil.h>
#include <ArmarXCore/core/logging/LoggingUtil.h>

#include <RobotAPI/libraries/armem/core/MemoryID.h>
#include <RobotAPI/libraries/armem/core/wm/memory_definitions.h>
#include <RobotAPI/libraries/armem/core/wm/aron_conversions.h>
#include <RobotAPI/libraries/armem/core/operations.h>
#include <RobotAPI/libraries/armem/server/wm/memory_definitions.h>


namespace armarx::armem::server::ltm
{
    /// @brief Interface functions for the longterm memory classes
    template <class _CoreSegmentT>
    class MemoryBase : public MemoryItem
    {
    public:
        using CoreSegmentT = _CoreSegmentT;

        MemoryBase() :
            MemoryBase({})
        {
        }

        MemoryBase(const MemoryID& id) :
            MemoryItem(id)
        {
        }

        /// return the full ltm as a wm::Memory with only references
        /// the ltm may be huge, use with caution
        armem::wm::Memory loadAllReferences()
        {
            armem::wm::Memory ret;
            loadAllReferences(ret);
            return ret;
        }
        void loadAllReferences(armem::wm::Memory& memory)
        {
            TIMING_START(LTM_Memory_LoadAll);
            _loadAllReferences(memory);
            TIMING_END_STREAM(LTM_Memory_LoadAll, ARMARX_DEBUG);
        }

        /// return the full ltm as a wm::Memory and resolves the references
        /// the ltm may be huge, use with caution
        armem::wm::Memory loadAllAndResolve()
        {
            armem::wm::Memory ret;
            loadAllAndResolve(ret);
            return ret;
        }

        void loadAllAndResolve(armem::wm::Memory& memory)
        {
            TIMING_START(LTM_Memory_LoadAllAndResolve);
            _loadAllReferences(memory);
            _resolve(memory);
            TIMING_END_STREAM(LTM_Memory_LoadAllAndResolve, ARMARX_DEBUG);
        }

        /// convert the references of the input into a wm::Memory
        void resolve(armem::wm::Memory& memory)
        {
            TIMING_START(LTM_Memory_Load);
            _resolve(memory);
            TIMING_END_STREAM(LTM_Memory_Load, ARMARX_DEBUG);
        }

        /// append a wm::Memory instance to the ltm
        void store(const armem::wm::Memory& memory)
        {
            TIMING_START(LTM_Memory_Append);
            for (auto& f : processors->memFilters)
            {
                if (f->enabled && !f->accept(memory))
                {
                    ARMARX_INFO << deactivateSpam() << "Ignoring to put a Memory into the LTM because it got filtered.";
                    return;
                }
            }
            _store(memory);
            TIMING_END_STREAM(LTM_Memory_Append, ARMARX_DEBUG);
        }

        /// append a wm::Memory instance to the ltm
        void store(const armem::server::wm::Memory& serverMemory)
        {
            wm::Memory memory;
            memory.update(armem::toCommit(serverMemory));
            this->store(memory);
        }

        /// iterate over all core segments of this ltm
        virtual bool forEachCoreSegment(std::function<void(CoreSegmentT&)>&& func) const = 0;

        /// find core segment
        virtual std::shared_ptr<CoreSegmentT> findCoreSegment(const std::string&) const = 0;

        /// parameters
        virtual void createPropertyDefinitions(PropertyDefinitionsPtr& defs, const std::string& prefix)
        {
            processors->createPropertyDefinitions(defs, prefix);
        }

        /// get level name
        static std::string getLevelName()
        {
            return "LT-Memory";
        }

    protected:
        virtual void _loadAllReferences(armem::wm::Memory& memory) = 0;
        virtual void _resolve(armem::wm::Memory& memory) = 0;
        virtual void _store(const armem::wm::Memory& memory) = 0;

    public:
        bool enabled = false;

    protected:
        mutable std::recursive_mutex ltm_mutex;

    };
} // namespace armarx::armem::server::ltm
