#include "Plugin.h"

#include <RobotAPI/libraries/armem/core/error.h>
#include <RobotAPI/libraries/armem/client/plugins/Plugin.h>

#include <ArmarXCore/core/Component.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>


namespace armarx::armem::server::plugins
{

    Plugin::~Plugin() = default;


    Plugin::Plugin(ManagedIceObject& parent, std::string prefix) :
        armarx::ComponentPlugin(parent, prefix)
    {
        addPlugin(clientPlugin);
        ARMARX_CHECK_NOT_NULL(clientPlugin);
        addPluginDependency(clientPlugin);
    }


    void Plugin::postCreatePropertyDefinitions(PropertyDefinitionsPtr& properties)
    {
        const std::string prefix = "mem.";
        if (not workingMemory.name().empty()
            and not properties->hasDefinition(prefix + "MemoryName"))
        {
            properties->optional(workingMemory.name(), prefix + "MemoryName", "Name of this memory server.");
        }

        // stuff for ltm
        if (not properties->hasDefinition(prefix + "ltm.enabled"))
        {
            properties->optional(longtermMemory.enabled, prefix + "ltm.enabled");
        }
        longtermMemory.createPropertyDefinitions(properties, prefix + "ltm.");

        // Publish memory updates topic
        properties->topic(memoryTopic, memoryTopicDefaultName);
    }


    void Plugin::postOnConnectComponent()
    {
        Component& parent = this->parent<Component>();

        // register to MNS
        if (clientPlugin->isMemoryNameSystemEnabled() and clientPlugin->getMemoryNameSystemClient())
        {
            registerServer(parent);
        }
        iceAdapter.setMemoryListener(memoryTopic);
    }


    void Plugin::preOnDisconnectComponent()
    {
        if (clientPlugin->isMemoryNameSystemEnabled() and clientPlugin->getMemoryNameSystemClient())
        {
            removeServer();
        }
    }


    void Plugin::setMemoryName(const std::string& memoryName)
    {
        workingMemory.name() = memoryName;
        longtermMemory.setMemoryID(workingMemory.id());
    }


    mns::dto::RegisterServerResult Plugin::registerServer(armarx::Component& parent)
    {
        MemoryID id = MemoryID().withMemoryName(workingMemory.name());
        mns::dto::MemoryServerInterfaces server;
        server.reading = ReadingMemoryInterfacePrx::uncheckedCast(parent.getProxy());
        server.writing = WritingMemoryInterfacePrx::uncheckedCast(parent.getProxy());
        server.actions = actions::ActionsInterfacePrx::uncheckedCast(parent.getProxy());

        mns::dto::RegisterServerResult result;
        try
        {
            clientPlugin->getMemoryNameSystemClient().registerServer(id, server);
            result.success = true;
            ARMARX_DEBUG << "Registered memory server for " << id << " in the Memory Name System (MNS).";
        }
        catch (const armem::error::ServerRegistrationOrRemovalFailed& e)
        {
            result.success = false;
            result.errorMessage = e.what();
            ARMARX_WARNING << e.what();
        }
        return result;
    }


    mns::dto::RemoveServerResult Plugin::removeServer()
    {
        MemoryID id = MemoryID().withMemoryName(workingMemory.name());

        mns::dto::RemoveServerResult result;
        try
        {
            clientPlugin->getMemoryNameSystemClient().removeServer(id);
            result.success = true;
            ARMARX_DEBUG << "Removed memory server for " << id << " from the Memory Name System (MNS).";
        }
        catch (const armem::error::ServerRegistrationOrRemovalFailed& e)
        {
            result.success = false;
            result.errorMessage = e.what();
            ARMARX_WARNING << e.what();
        }
        catch (const Ice::NotRegisteredException&)
        {
            // It's ok, the MNS is gone.
            result.success = false;
            result.errorMessage = "Memory Name System is gone.";
        }
        return result;
    }

}
